ul.todo_list {
    list-style-type: none;
    margin: 0;
    padding: 0;
}

ul.todo_list li {
    display: block;
    margin: 0;
    padding: 7px 0;
    border-top: 1px solid #eee;
}

ul.todo_list li p {
    display: inline;
}

ul.todo_list li p.link {
    font-weight: bold;
}

ul.todo_list li p.details {
    font-style: italic;
}

ul.todo_list li {
}

div.admonition {
    border: 1px solid #8F1000;
}

div.admonition p.admonition-title {
    background-color: #8F1000;
    border-bottom: 1px solid #8E8E8E;
}

a {
    color: #CF2F19;
}

div.related ul li a {
    color: #CF2F19;
}

div.sphinxsidebar h4 {
    background-color:#8E8E8E;
    border:1px solid #255E6E;
    color:white;
    font-size:1em;
    margin:1em 0 0.5em;
    padding:0.1em 0 0.1em 0.5em;
}

em {
    font-style: normal;
}

table.docutils {
    font-size: 11px;
}

.tweet_list li {
    font-size: 0.9em;
    border-bottom: 1px solid #eee;
    padding: 5px 0;
}

.tweet_list li .tweet_avatar {
    float: left;
}

/* ------------------------------------------
PURE CSS SPEECH BUBBLES
by Nicolas Gallagher
- http://nicolasgallagher.com/pure-css-speech-bubbles/

http://nicolasgallagher.com
http://twitter.com/necolas

Created: 02 March 2010
Version: 1.1 (21 October 2010)

Dual licensed under MIT and GNU GPLv2 © Nicolas Gallagher
------------------------------------------ */
/* THE SPEECH BUBBLE
------------------------------------------------------------------------------------------------------------------------------- */

/* THE SPEECH BUBBLE
------------------------------------------------------------------------------------------------------------------------------- */

.triangle-border {
    position:relative;
    padding:15px;
    margin:1em 0 3em;
    border:5px solid #BC1518;
    color:#333;
    background:#fff;

    /* css3 */
    -moz-border-radius:10px;
    -webkit-border-radius:10px;
    border-radius:10px;
}

/* Variant : for left positioned triangle
------------------------------------------ */

.triangle-border.left {
    margin-left:30px;
}

/* Variant : for right positioned triangle
------------------------------------------ */

.triangle-border.right {
    margin-right:30px;
}

/* THE TRIANGLE
------------------------------------------------------------------------------------------------------------------------------- */

.triangle-border:before {
    content:"";
    display:block; /* reduce the damage in FF3.0 */
    position:absolute;
    bottom:-40px; /* value = - border-top-width - border-bottom-width */
    left:40px; /* controls horizontal position */
    width:0;
    height:0;
    border:20px solid transparent;
    border-top-color:#BC1518;
}

/* creates the smaller  triangle */
.triangle-border:after {
    content:"";
    display:block; /* reduce the damage in FF3.0 */
    position:absolute;
    bottom:-26px; /* value = - border-top-width - border-bottom-width */
    left:47px; /* value = (:before left) + (:before border-left) - (:after border-left) */
    width:0;
    height:0;
    border:13px solid transparent;
    border-top-color:#fff;
}

/* Variant : top
------------------------------------------ */

/* creates the larger triangle */
.triangle-border.top:before {
    top:-40px; /* value = - border-top-width - border-bottom-width */
    right:40px; /* controls horizontal position */
    bottom:auto;
    left:auto;
    border:20px solid transparent;
    border-bottom-color:#BC1518;
}

/* creates the smaller  triangle */
.triangle-border.top:after {
    top:-26px; /* value = - border-top-width - border-bottom-width */
    right:47px; /* value = (:before right) + (:before border-right) - (:after border-right) */
    bottom:auto;
    left:auto;
    border:13px solid transparent;
    border-bottom-color:#fff;
}

/* Variant : left
------------------------------------------ */

/* creates the larger triangle */
.triangle-border.left:before {
    top:10px; /* controls vertical position */
    left:-30px; /* value = - border-left-width - border-right-width */
    bottom:auto;
    border-width:15px 30px 15px 0;
    border-style:solid;
    border-color:transparent #BC1518;
}

/* creates the smaller  triangle */
.triangle-border.left:after {
    top:16px; /* value = (:before top) + (:before border-top) - (:after border-top) */
    left:-21px; /* value = - border-left-width - border-right-width */
    bottom:auto;
    border-width:9px 21px 9px 0;
    border-style:solid;
    border-color:transparent #fff;
}

/* Variant : right
------------------------------------------ */

/* creates the larger triangle */
.triangle-border.right:before {
    top:10px; /* controls vertical position */
    right:-30px; /* value = - border-left-width - border-right-width */
    bottom:auto;
    left:auto;
    border-width:15px 0 15px 30px;
    border-style:solid;
    border-color:transparent #BC1518;
}

/* creates the smaller  triangle */
.triangle-border.right:after {
    top:16px; /* value = (:before top) + (:before border-top) - (:after border-top) */
    right:-21px; /* value = - border-left-width - border-right-width */
    bottom:auto;
    left:auto;
    border-width:9px 0 9px 21px;
    border-style:solid;
    border-color:transparent #fff;
}

