/*
 * (C) 2009 Rackspace Hosting, All Rights Reserved.
 */
body, div, dl, dt, dd, ul, ol, li, h2, h3,
h4, h5, h6, pre, code, form, fieldset, legend,
input, button, textarea, p, blockquote, th, td {
    text-align: left;
}

h1 {
    font-size: 350%;
    margin-bottom: 10px;
}

#Content {
    border: 1px solid;
    padding: 0px 40px 40px;
    margin-left: 155px;
}

#SrcContent {
    padding: 0px 40px 40px;
    display: none;
    margin-left: 155px;
}

#Controller {
    position: fixed;
    width: 145px;
    left: 10px;
    top: 10px;
}

.Sample {
    display: none;
}

.EnumValue{
    padding: 10px 0px;
}

.EnumDoc{
    padding: 10px 10px 10px 0px;
}

.ExternHref{
    padding-top: 5px;
}

.ExternDoc{
    padding-right: 10px;
}

pre {
    overflow: auto;
}

td {
    padding: 0px 0px 0px 10px;
    width: 50%;
    font-size: 90%;
}

table {
    width: 100%;
}

a {
 text-decoration: none;
}

a:hover {
    text-decoration: underline;
}

a:link {
    color: #000090;
}

a:visited {
    color: #000090;
}
