# Copyright (c) 2014 OpenStack Foundation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo.config import cfg

from neutron.extensions import portbindings

eswitch_opts = [
    cfg.StrOpt('vnic_type',
               default=portbindings.VIF_TYPE_MLNX_DIRECT,
               help=_("Type of VM network interface: mlnx_direct or "
                      "hostdev")),
    cfg.BoolOpt('apply_profile_patch',
                default=False,
                help=_("Enable server compatibility with old nova")),
]


cfg.CONF.register_opts(eswitch_opts, "ESWITCH")
